"""SkeleParser"""

import sys
import argparse
from ...common import DESCRIPTION
from ...common import VERSION

SCF_ARG = "scaffold"
SCF_HELP = "Scaffold a new or existing project with Skelebot"
SCF_EX = "--existing"
SCF_EX_ALT = "-e"
SCF_EX_HELP = "Scaffold an existing project without creating new folder"

ENV_ARG = "--env"
ENV_ALT = "-e"
ENV_HELP = "Specify the runtime environment configurations"

SB_DST = "skip_build_global"
SB_ARG = "--skip-build"
SB_ALT = "-s"
SB_HELP = "Skip the build process and attempt to use previous docker build"

HS_DST = "host"
HS_ARG = "--docker-host"
HS_ALT = "-d"
HS_HELP = "Set the Docker Host on which the command will be executed"

NT_DST = "native_global"
NT_ARG = "--native"
NT_ALT = "-n"
NT_HELP = "Run natively instead of through Docker"

VN_DST = "version_global"
VN_ARG = "--version"
VN_ALT = "-v"
VN_HELP = "Display the version number of Skelebot"

CN_DST = "contact_global"
CN_ARG = "--contact"
CN_ALT = "-c"
CN_HELP = "Display the contact email of the Skelebot project"

def addArgs(args, subparser):
    """Add args to the given subparser"""

    if (args is not None):
        for arg in args:
            if not arg.choices:
                subparser.add_argument(arg.name, help=arg.help)
            else:
                subparser.add_argument(arg.name, choices=arg.choices, help=arg.help)
    return subparser

def addParams(params, subparser):
    """Add params to the given subparser"""

    if (params is not None):
        for param in params:
            flags = ["--{}".format(param.name)]
            if (param.alt is not None):
                flags.append("-{}".format(param.alt))

            if param.choices:
                subparser.add_argument(*flags, choices=param.choices, default=param.default,
                                       help=param.help)
            elif param.accepts == "boolean":
                subparser.add_argument(*flags, action='store_true', help=param.help)
            elif param.accepts == "list":
                subparser.add_argument(*flags, nargs='*', default=param.default,
                                       help=param.help)
            else:
                subparser.add_argument(*flags, default=param.default, help=param.help)
    return subparser

class SkeleParser:
    """
    SkeleParser Class

    A wrapper around the argparse that constructs the interface to Skelebot based on the user's
    skelebot.yaml configuration file
    """

    parser = None
    config = None
    env = None
    desc = None

    # Initialize the parser with the given config and environment
    def __init__(self, config=None, env=None):
        """Initialize the argparse Parser based on the Config data if it is present"""

        self.config = config
        self.env = env
        self.desc = self.buildDescription()

        # ---Standard Parser Setup---

        # Construct the root argument parser from which all sub-parsers will be built
        formatter = argparse.RawTextHelpFormatter
        self.parser = argparse.ArgumentParser(description=self.desc, formatter_class=formatter)
        self.parser.add_argument(VN_ALT, VN_ARG, help=VN_HELP, action='store_true', dest=VN_DST)
        subparsers = self.parser.add_subparsers(dest="job")

        if (config.name is None):
            # Add SCF parser
            scaffoldParser = subparsers.add_parser(SCF_ARG, help=SCF_HELP)
            scaffoldParser.add_argument(SCF_EX_ALT, SCF_EX, action='store_true', help=SCF_EX_HELP)
        else:
            # Add STANDARD PARAMS
            self.parser.add_argument(ENV_ALT, ENV_ARG, help=ENV_HELP)
            self.parser.add_argument(HS_ALT, HS_ARG, help=HS_HELP, dest=HS_DST)
            self.parser.add_argument(SB_ALT, SB_ARG, help=SB_HELP, action='store_true', dest=SB_DST)
            self.parser.add_argument(NT_ALT, NT_ARG, help=NT_HELP, action='store_true', dest=NT_DST)
            self.parser.add_argument(CN_ALT, CN_ARG, help=CN_HELP, action='store_true', dest=CN_DST)

        # ---Config Based Parser Setup---

        # Add JOBS
        if (config.jobs is not None):
            for job in config.jobs:
                subparser = subparsers.add_parser(job.name, help=job.help + " ("+job.source+")")

                # Add ARGS and PARAMS
                subparser = addArgs(job.args, subparser)
                subparser = addParams(job.params, subparser)
                subparser = addParams(config.params, subparser)

        # Add COMPONENT parsers
        for component in self.config.components:
            subparsers = component.addParsers(subparsers)

    def parseArgs(self, args=None):
        """Parse the args from the parser built by the config"""
        return self.parser.parse_args(args if args is not None else sys.argv[1:])

    def showHelp(self):
        """Display the help message from the internal parser object"""
        self.parser.print_help()

    def buildDescription(self):
        """Construct the description text for the '--help' output"""

        description = "Skelebot Version: {version}".format(version=VERSION)
        if (self.config.name is not None):
            name = " ".join([word.capitalize() for word in self.config.name.split("-")])
            description = self.config.description
            version = self.config.version
            description = DESCRIPTION.format(version=VERSION, project=name, desc=description,
                                             pVersion=version, env=self.env)

        return description



import copy
import unittest

from schema import SchemaError

import skelebot as sb

class TestJobValidate(unittest.TestCase):

    job = {
        'name': 'test',
        'source': 'test',
        'mode': 'test',
        'help': 'test',
        'args': [1, 2],
        'params': [1, 2],
        'ignores': [1, 2],
        'mappings': [1, 2]
    }

    def validate_error(self, attr, reset, expected):
        job = copy.deepcopy(self.job)
        job[attr] = reset

        try:
            sb.objects.job.Job.validate(job)
            self.fail("Exception Expected")
        except SchemaError as error:
            self.assertEqual(error.code, "Job '{attr}' must be a {expected}".format(attr=attr, expected=expected))

    def test_valid(self):

        try:
            sb.objects.job.Job.validate(self.job)
        except:
            self.fail("Validation Raised Exception Unexpectedly")

    def test_invalid_mising(self):
        job = copy.deepcopy(self.job)
        del job['name']
        del job['source']
        del job['help']

        try:
            sb.objects.job.Job.validate(job)
            self.fail("Exception Expected")
        except SchemaError as error:
            self.assertEqual(error.code, "Missing keys: 'help', 'name', 'source'")

    def test_invalid(self):
        self.validate_error('name', 123, 'String')
        self.validate_error('source', 123, 'String')
        self.validate_error('mode', 123, 'String')
        self.validate_error('help', 123, 'String')
        self.validate_error('args', 123, 'List')
        self.validate_error('params', 123, 'List')
        self.validate_error('ignores', 123, 'List')
        self.validate_error('mappings', 123, 'List')

if __name__ == '__main__':
    unittest.main()



import os
import unittest
from unittest import mock

import skelebot as sb

class TestREADME(unittest.TestCase):
    path = ""
    config = None

    # Get the path to the current working directory before we mock the function to do so
    def setUp(self):
        self.path = os.getcwd()

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildREADME(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/README.md".format(folder=folderPath)
        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        self.config = sb.systems.generators.yaml.loadConfig()

        expected = """# test
![Version](https://img.shields.io/badge/Version-6.6.6-brightgreen.svg)
![Documentation](https://img.shields.io/badge/Documentation-UNLINKED-red.svg)

test cases

---

## Contact
Project Maintainer: Mega Man (megaman@cars.com)"""


        sb.systems.generators.readme.buildREADME(self.config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expected)

if __name__ == '__main__':
    unittest.main()



"""Job Class"""

from schema import Schema, And, Optional
from .skeleYaml import SkeleYaml
from .param import Param
from .arg import Arg

class Job(SkeleYaml):
    """
    Job Class

    Config class that defines a job in a Skelebot project via the config yaml file
    """

    schema = Schema({
        'name': And(str, error='Job \'name\' must be a String'),
        'source': And(str, error='Job \'source\' must be a String'),
        Optional('mode'): And(str, error='Job \'mode\' must be a String'),
        Optional('host'): And(str, error='Job \'host\' must be a String'),
        'help': And(str, error='Job \'help\' must be a String'),
        Optional('args'): And(list, error='Job \'args\' must be a List'),
        Optional('params'): And(list, error='Job \'params\' must be a List'),
        Optional('ignores'): And(list, error='Job \'ignores\' must be a List'),
        Optional('mappings'): And(list, error='Job \'mappings\' must be a List'),
        Optional('ports'): And(list, error='\'ports\' must be a List')
    }, ignore_extra_keys=True)

    name = None
    source = None
    mode = None
    host = None
    help = None
    args = None
    params = None
    ignores = None
    mappings = None
    ports = None

    def __init__(self, name=None, source=None, mode=None, host=None, help=None, args=None,
                 params=None, ignores=None, mappings=None, ports=None):
        """Initialize the job object with all provided optional attributes"""

        self.name = name
        self.source = source
        self.help = help
        self.mode = mode if mode is not None else "i"
        self.host = host
        self.args = args if args is not None else []
        self.params = params if params is not None else []
        self.ignores = ignores if ignores is not None else []
        self.mappings = mappings if mappings is not None else []
        self.ports = ports if ports is not None else []

    @classmethod
    def load(cls, config):
        """
        Defines the customer manner in which the object is loaded from a Dict due to the nature of
        args and params needing to be loaded differently since they are lists of objects that need
        to be instantiated
        """

        cls.validate(config)

        values = {}
        for attr in config.keys():
            if attr in list(vars(Job).keys()):
                if attr == "args":
                    values[attr] = Arg.loadList(config[attr])
                elif attr == "params":
                    values[attr] = Param.loadList(config[attr])
                else:
                    values[attr] = config[attr]

        return cls(**values)



import artifactory
import shutil
import os
from schema import Schema, And
from .artifactRepo import ArtifactRepo
from ...objects.semver import Semver

ARTIFACT_PATH = "{url}/{repo}/{path}/{file}"

class ArtifactoryRepo(ArtifactRepo):
    """
    Artifactory Repository Class

    Contains details and functions for pushing and pulling artifacts from Artifactory
    """

    schema = Schema({
        'url': And(str, error="Artifactory 'url' must be a String"),
        'repo': And(str, error="Artifactory 'repo' must be a String"),
        'path': And(str, error="Artifactory 'path' must be a String"),
    }, ignore_extra_keys=True)

    url = None
    repo = None
    path = None

    def __init__(self, url, repo, path):
        """ Initialize the values for an Artifactory connection and set the global config """
        self.url = url
        self.repo = repo
        self.path = path

        artifactory.global_config = {
            self.url: {
                'username': None,
                'verify': True,
                'cert': None,
                'password': None
            }
        }

    def getUsername(self):
        """ Prompt for an Artifactory username """
        return input("Please provide a valid Artifactory user: ")

    def getToken(self):
        """ Prompt for an Artifactory token """
        return input("Please provide a valid Artifactory token: ")

    def getLatestCompatibleVersion(self, artifact, currentVersion, user, token):
        """ Search for the latest compatible version for the provided version number """

        print("Searching for Latest Compatible Artifact")
        version = None
        currentSemver = Semver(currentVersion)

        listUrl = "{url}/{repo}/{path}/".format(url=self.url, repo=self.repo, path=self.path)
        ext = artifact.file.split(".")[1]
        path = artifactory.ArtifactoryPath(listUrl, auth=(user, token))

        # Iterate over all artifacts in the ArtifactoryPath (because path.glob was throwing exceptions on Linux systems)
        if (path.exists()):
            for pathArtifact in path: # Only look at artifacts with the same filename and major version
                modelPrefix = "{filename}_v{major}".format(filename=artifact.name, major=currentSemver.major)
                if modelPrefix in str(pathArtifact):
                    artifactSemver = Semver(str(pathArtifact).split("_v")[1].split(ext)[0])
                    if (currentSemver.isBackwardCompatible(artifactSemver)) and ((version is None) or (version < artifactSemver)):
                        version = artifactSemver # Identify the latest compatible version

        # Raise an error if no compatible version is found
        if (version is None):
            raise RuntimeError(self.ERROR_NOT_COMPATIBLE)

        return version

    def push(self, artifact, version, force=False, user=None, password=None):
        """ Push the artifact with the given version number to Artifactory """

        # Obtain any missing credentials
        if (user is None):
            user = self.getUsername()
        if (password is None):
            password = self.getToken()

        artifactName = artifact.getVersionedName(version)
        url = ARTIFACT_PATH.format(url=self.url, repo=self.repo, path=self.path, file=artifactName)

        # Error and exit if artifact already exists and we are not forcing an override
        if (not force) and (artifactory.ArtifactoryPath(url, auth=(user, password)).exists()):
            raise RuntimeError(self.ERROR_ALREADY_PUSHED)

        # Push the artifact to the path in Artifactory
        print("Deploying {file} to {url}".format(file=artifactName, url=url))
        shutil.copyfile(artifact.file, artifactName)
        path = artifactory.ArtifactoryPath(url, auth=(user, password))
        try:
            path.deploy_file(artifactName)
            os.remove(artifactName)
        except:
            os.remove(artifactName)
            raise

    def pull(self, artifact, version, currentVersion=None, override=False, user=None, password=None):
        """ Pull the artifact with the given version number (or LATEST) from Artifactory """

        # Obtain any missing credentials
        if (user is None):
            user = self.getUsername()
        if (password is None):
            password = self.getToken()

        # Find Latest Compatible Version or Use Provided Version Number if the artifact is not 'singular'
        version = version if (version != "LATEST" or artifact.singular == True) else self.getLatestCompatibleVersion(artifact, currentVersion, user, password)

        # Construct the Artifactory URL for the specified Artifact
        artifactName = artifact.getVersionedName(version)
        url = ARTIFACT_PATH.format(url=self.url, repo=self.repo, path=self.path, file=artifactName)

        # Download Artifact from Artifactory and save it to the disk
        if (artifactory.ArtifactoryPath(url, auth=(user, password)).exists()):
            print("Pulling {file} from {url}".format(file=artifactName, url=url))
            path = artifactory.ArtifactoryPath(url, auth=(user, password))
            with path.open() as fd:
                dest = artifact.file if (override) else artifactName
                with open(dest, "wb") as out:
                    out.write(fd.read())
        else:
            print("Artifact Not Found: {url}".format(url=url))



"""Bump Component"""

from ..objects.component import Activation, Component
from ..systems.generators import yaml

MAJOR = "major"
MINOR = "minor"
PATCH = "patch"

class Bump(Component):
    """
    Bump Class

    Provides the ability to bump the major, minor, or patch version of a project based on Semantic
    Versioning standards with the 'bump' command
    """
    activation = Activation.PROJECT
    commands = ["bump"]

    def addParsers(self, subparsers):
        """
        SkeleParser Hook

        Adds a parser for the bump command that takes in a single argument called version to define
        whether it should bump the major, minor, or patch version of the project
        """

        helpMessage = "Bump the skelebot.yaml project version"
        argHelp = "Select the version number that should be bumped"

        parser = subparsers.add_parser("bump", help=helpMessage)
        parser.add_argument("version", help=argHelp, choices=[MAJOR, MINOR, PATCH])

        return subparsers

    def execute(self, config, args, host=None):
        """
        Execution Hook

        Executes when the bump command is provided and bumps either the major, minor, or patch
        version of the project depending on which was provided in the command
        """

        oldVersion = config.version
        version = args.version
        mmp = oldVersion.split(".")
        mmp[0] = str(int(mmp[0])+1) if version == MAJOR else mmp[0]
        mmp[1] = str(int(mmp[1])+1) if version == MINOR else (mmp[1] if version == PATCH else "0")
        mmp[2] = str(int(mmp[2])+1) if version == PATCH else "0"
        version = ".".join(mmp)
        yaml.saveVersion(version)

        print("Bumped " + version + " version. v" + oldVersion + " -> v" + version)



"""Root Config Class for Skelebot YAML File"""

from schema import Schema, And, Or, Use, Optional
from .job import Job
from .param import Param
from .skeleYaml import SkeleYaml
from .component import Activation
from ..common import LANGUAGE_IMAGE, PYTHON_VERSIONS
from ..components.componentFactory import ComponentFactory

class Config(SkeleYaml):
    """
    Root Config Class for Skelebot YAML File

    Built on top of the SkeleYaml parent Object in order to enherit and extend the functionality
    of yaml file generation and parsing
    """

    schema = Schema({
        'name': And(str, error='\'name\' must be a String'),
        Optional('env'): And(str, error='\'env\' must be a String'),
        Optional('description'): And(str, error='\'description\' must be a String'),
        Optional('maintainer'): And(str, error='\'maintainer\' must be a String'),
        Optional('contact'): And(str, error='\'contact\' must be a String'),
        Optional('host'): And(str, error='\'host\' must be a String'),
        'language': And(str, error='\'language\' must be a String'),
        Optional('baseImage'): And(str, error='\'baseImage\' must be a String'),
        Optional('timezone'): And(str, error='\'timezone\' must be a String'),
        Optional('primaryJob'): And(str, error='\'primaryJob\' must be a String'),
        Optional('primaryExe'): And(str, Use(str.upper), lambda s: s in ('CMD', 'ENTRYPOINT'), error='\'primaryExe\' must be CMD or ENTRYPOINT'),
        Optional('ephemeral'): And(bool, error='\'ephemeral\' must be a Boolean'),
        Optional('dependencies'): Or(dict, list, error='\'dependencies\' must be a Dict or List'),
        Optional('ignores'): And(list, error='\'ignores\' must be a List'),
        Optional('jobs'): And(list, error='\'jobs\' must be a List'),
        Optional('ports'): And(list, error='\'ports\' must be a List'),
        Optional('components'): And(dict, error='\'components\' must be a Dictionary'),
        Optional('params'): And(list, error='\'params\' must be a List'),
        Optional('commands'): And(list, error='\'commands\' must be a List'),
        Optional('pythonVersion'): And(str, Or(*PYTHON_VERSIONS), error='\'pythonVersion\' must be one of:' + ', '.join(PYTHON_VERSIONS))
    }, ignore_extra_keys=True)

    name = None
    env = None
    description = None
    version = None
    maintainer = None
    contact = None
    host = None
    language = None
    baseImage = None
    timezone = None
    primaryJob = None
    primaryExe = None
    ephemeral = None
    dependencies = None
    ignores = None
    jobs = None
    ports = None
    components = None
    params = None
    commands = None
    pythonVersion = '3.6'

    def __init__(self, name=None, env=None, description=None, version=None, maintainer=None,
                 contact=None, host=None, language=None, baseImage=None, timezone=None,
                 primaryJob=None, primaryExe=None, ephemeral=None, dependencies=None, ignores=None,
                 jobs=None, ports=None, components=None, params=None, commands=None, pythonVersion = '3.6'):
        """Initialize the config object with all provided optional attributes"""

        self.name = name
        self.env = env
        self.description = description
        self.version = version
        self.maintainer = maintainer
        self.contact = contact
        self.host = host
        self.language = language
        self.baseImage = baseImage
        self.timezone = timezone
        self.primaryJob = primaryJob
        self.primaryExe = primaryExe.upper() if primaryExe is not None else "CMD"
        self.ephemeral = ephemeral
        self.dependencies = dependencies if dependencies is not None else []
        self.ignores = ignores if ignores is not None else []
        self.jobs = jobs if jobs is not None else []
        self.ports = ports if ports is not None else []
        self.components = components if components is not None else []
        self.params = params if params is not None else []
        self.commands = commands if commands is not None else []
        self.pythonVersion = pythonVersion

    def toDict(self):
        """
        Extends the parent function in order to logic for handling the conversion of
        components since the Class structure and yaml sctructures do not match
        """

        components_dict = {}
        for component in self.components:
            component_dict = component.toDict()
            if component_dict != {}:
                components_dict[component.__class__.__name__.lower()] = component_dict

        dct = super().toDict()
        dct["components"] = components_dict
        return dct

    def getBaseImage(self):
        """
        Returns the proper base image based on the values for language and kerberos,
        or returns the user defined base image if provided in the config yaml
        """

        if self.baseImage:
            image = self.baseImage
        else:
            language = self.language if self.language is not None else "NA"
            image = LANGUAGE_IMAGE[language]
            if language == 'Python':
                image['base'] = image['base'].format(pythonVersion = self.pythonVersion)

            variant = "base"
            for component in self.components:
                if component.__class__.__name__.lower() == "kerberos":
                    variant = "krb"

            image = image[variant]

        return image

    def getImageName(self):
        """Construct and return the name for the docker image based on the project name"""
        image_name = self.name.lower().replace(" ", "-")
        if self.env:
            image_name += "-{env}".format(env=self.env)
        return image_name

    def getHost(self, job=None, args=None):
        host = self.host if self.host is not None else None
        if job is not None:
            host = job.host if job.host is not None else host
        if args is not None:
            host = args.host if getattr(args, 'host', None) is not None else host
        return host

    def loadComponents(self, config):
        """
        Parses the components section of skelebot.yaml config to generate the complete list of
        components for the project based on the active component list and each components'
        Activation attribute
        """

        componentFactory = ComponentFactory()
        if (config is None):
            # Build the default components for a non-skelebot project
            defaultActivations = [Activation.EMPTY, Activation.ALWAYS]
            self.components = componentFactory.buildComponents(defaultActivations)
        else:
            # Build the components that are defined in the config yaml data
            compNames = []
            components = []
            if ("components" in config):
                configComps = config["components"]
                for compName in configComps:
                    # Ensures that artifactory component is not loaded if repository is present
                    if ("repository" not in configComps) or (compName != "artifactory"):
                        component = componentFactory.buildComponent(compName, configComps[compName])
                        if (component is not None):
                            components.append(component)
                            compNames.append(component.__class__.__name__)

            # Build the additonal components that are active without configuration data
            activations = [Activation.PROJECT, Activation.ALWAYS]
            components.extend(componentFactory.buildComponents(activations, ignores=compNames))
            self.components = components

    @classmethod
    def load(cls, config):
        """Load the config Dict from the yaml file into the Config object"""

        cfg = cls()
        if config is not None:

            cls.validate(config)
            values = {}
            for attr, value in config.items():
                if (attr in vars(Config)) and (attr != "components") and (attr != "version"):
                    if (attr == "jobs"):
                        values[attr] = Job.loadList(value)
                    elif (attr == "params"):
                        values[attr] = Param.loadList(value)
                    else:
                        values[attr] = value

            cfg = cls(**values)

        cfg.loadComponents(config)

        return cfg



import os
import unittest
from unittest import mock

import skelebot as sb

# Test plugin that says 'Duuuuuude' at the end of every command
class SayDude(sb.objects.component.Component):
    activation = sb.objects.component.Activation.ALWAYS

    def appendCommand(self, job, native):
        return "echo Duuuuuude"

class TestDockerfile(unittest.TestCase):
    path = ""

    # Get the path to the current working directory before we mock the function to do so
    def setUp(self):
        self.path = os.getcwd()
        self.maxDiff = None

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_no_language(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)
        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = None

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM ubuntu:18.04
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_base(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "R"
        config.dependencies.append("github:github.com/repo:cool-lib")
        config.dependencies.append("file:libs/proj:cool-proj")
        config.dependencies.append("dtable=9.0")

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/r-base
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["Rscript", "-e", "install.packages('pyyaml', repo='https://cloud.r-project.org'); library(pyyaml)"]
RUN ["Rscript", "-e", "install.packages('artifactory', repo='https://cloud.r-project.org'); library(artifactory)"]
RUN ["Rscript", "-e", "install.packages('argparse', repo='https://cloud.r-project.org'); library(argparse)"]
RUN ["Rscript", "-e", "install.packages('coverage', repo='https://cloud.r-project.org'); library(coverage)"]
RUN ["Rscript", "-e", "install.packages('pytest', repo='https://cloud.r-project.org'); library(pytest)"]
RUN ["Rscript", "-e", "library(devtools); install_github('github.com/repo'); library(cool-lib)"]
COPY libs/proj libs/proj
RUN ["Rscript", "-e", "install.packages('/app/libs/proj', repos=NULL, type='source'); library(cool-proj)"]
RUN ["Rscript", "-e", "library(devtools); install_version('dtable', version='9.0', repos='http://cran.us.r-project.org'); library(dtable)"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_entrypoint_exec(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.primaryExe = "ENTRYPOINT"

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/python-base:3.6
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
ENTRYPOINT [\"bash\", \"build.sh\"]\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_entrypoint_path(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.primaryExe = "ENTRYPOINT"

        config.primaryJob = "test-entrypoint-path"
        job = sb.objects.job.Job(name="test-entrypoint-path", source="jobs/dummy.py")
        config.jobs.append(job)

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/python-base:3.6
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
ENTRYPOINT ["python", "-u", "jobs/dummy.py"]\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_cmd_path(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.primaryExe = "CMD"

        config.primaryJob = "test-cmd-path"
        job = sb.objects.job.Job(name="test-cmd-path", source="jobs/dummy.py")
        config.jobs.append(job)

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/python-base:3.6
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c "python -u jobs/dummy.py --log info"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_base_py(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "Python"
        config.dependencies.append("github:github.com/repo")
        config.dependencies.append("github:https://github.com/securerepo")
        config.dependencies.append("file:libs/proj")
        config.dependencies.append("dtable=9.0")

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/python-base:3.6
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
RUN ["pip", "install", "git+github.com/repo"]
RUN ["pip", "install", "git+https://github.com/securerepo"]
COPY libs/proj libs/proj
RUN ["pip", "install", "/app/libs/proj"]
RUN ["pip", "install", "dtable==9.0"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_base_py_versions(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "Python"
        config.dependencies.append("dtable=9.0")
        config.dependencies.append("pandas==0.25")
        config.dependencies.append("numpy~=1.17")
        config.dependencies.append("requests>= 2.2, == 2.*")
        config.dependencies.append("scipy!= 1.3.*")

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/python-base:3.6
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
RUN ["pip", "install", "dtable==9.0"]
RUN ["pip", "install", "pandas==0.25"]
RUN ["pip", "install", "numpy~=1.17"]
RUN ["pip", "install", "requests>= 2.2, == 2.*"]
RUN ["pip", "install", "scipy!= 1.3.*"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_krb(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "R"
        config.dependencies.append("github:github.com/repo:cool-lib")
        config.dependencies.append("file:libs/proj:cool-proj")
        config.dependencies.append("dtable=9.0")
        config.components.append(sb.components.kerberos.Kerberos("conf", "tab", "user"))

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/r-krb
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["Rscript", "-e", "install.packages('pyyaml', repo='https://cloud.r-project.org'); library(pyyaml)"]
RUN ["Rscript", "-e", "install.packages('artifactory', repo='https://cloud.r-project.org'); library(artifactory)"]
RUN ["Rscript", "-e", "install.packages('argparse', repo='https://cloud.r-project.org'); library(argparse)"]
RUN ["Rscript", "-e", "install.packages('coverage', repo='https://cloud.r-project.org'); library(coverage)"]
RUN ["Rscript", "-e", "install.packages('pytest', repo='https://cloud.r-project.org'); library(pytest)"]
RUN ["Rscript", "-e", "library(devtools); install_github('github.com/repo'); library(cool-lib)"]
COPY libs/proj libs/proj
RUN ["Rscript", "-e", "install.packages('/app/libs/proj', repos=NULL, type='source'); library(cool-proj)"]
RUN ["Rscript", "-e", "library(devtools); install_version('dtable', version='9.0', repos='http://cran.us.r-project.org'); library(dtable)"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
COPY conf /etc/krb5.conf
COPY tab /krb/auth.keytab
CMD /bin/bash -c \"/./krb/init.sh user && bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_no_command(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        # No custom docker run command
        config.commands = []
        config.language = None

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM ubuntu:18.04
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
COPY . /app
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_append_command(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = None
        config.components.append(SayDude())

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM ubuntu:18.04
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info && echo Duuuuuude\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_custom(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.baseImage = "whatever:uwant"

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM whatever:uwant
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip", "install", "pyyaml"]
RUN ["pip", "install", "artifactory"]
RUN ["pip", "install", "argparse"]
RUN ["pip", "install", "coverage"]
RUN ["pip", "install", "pytest"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_R_plus_Python(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "R+Python"
        config.dependencies = {
            "Python":[
                "numpy", "pandas",
                "github:github.com/repo", "github:https://github.com/securerepo",
                "file:libs/proj",
                "dtable>=9.0", "dtable=9.0"
            ],
            "R":[
                "data.table", "here",
                "github:github.com/repo:cool-lib",
                "file:libs/proj:cool-proj",
                "dtable=9.0"
            ]
        }
        config.components.append(sb.components.kerberos.Kerberos("conf", "tab", "user"))

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/r-krb
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
RUN ["pip3", "install", "numpy"]
RUN ["pip3", "install", "pandas"]
RUN ["pip3", "install", "git+github.com/repo"]
RUN ["pip3", "install", "git+https://github.com/securerepo"]
COPY libs/proj libs/proj
RUN ["pip3", "install", "/app/libs/proj"]
RUN ["pip3", "install", "dtable>=9.0"]
RUN ["pip3", "install", "dtable==9.0"]
RUN ["Rscript", "-e", "install.packages('data.table', repo='https://cloud.r-project.org'); library(data.table)"]
RUN ["Rscript", "-e", "install.packages('here', repo='https://cloud.r-project.org'); library(here)"]
RUN ["Rscript", "-e", "library(devtools); install_github('github.com/repo'); library(cool-lib)"]
COPY libs/proj libs/proj
RUN ["Rscript", "-e", "install.packages('/app/libs/proj', repos=NULL, type='source'); library(cool-proj)"]
RUN ["Rscript", "-e", "library(devtools); install_version('dtable', version='9.0', repos='http://cran.us.r-project.org'); library(dtable)"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
COPY conf /etc/krb5.conf
COPY tab /krb/auth.keytab
CMD /bin/bash -c "/./krb/init.sh user && bash build.sh --env local --log info\"\n"""
        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerfile_timezone(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/Dockerfile".format(folder=folderPath)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        config = sb.systems.generators.yaml.loadConfig()
        config.language = "R"
        config.timezone = "America/Chicago"
        config.dependencies.append("github:github.com/repo:cool-lib")
        config.dependencies.append("file:libs/proj:cool-proj")
        config.dependencies.append("dtable=9.0")

        expectedDockerfile = """
# This Dockerfile was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

FROM skelebot/r-base
MAINTAINER Mega Man <megaman@cars.com>
WORKDIR /app
ENV TZ=America/Chicago
RUN ln -snf /usr/share/zoneinfo/$TZ /etc/localtime && echo $TZ > /etc/timezone
RUN ["Rscript", "-e", "install.packages('pyyaml', repo='https://cloud.r-project.org'); library(pyyaml)"]
RUN ["Rscript", "-e", "install.packages('artifactory', repo='https://cloud.r-project.org'); library(artifactory)"]
RUN ["Rscript", "-e", "install.packages('argparse', repo='https://cloud.r-project.org'); library(argparse)"]
RUN ["Rscript", "-e", "install.packages('coverage', repo='https://cloud.r-project.org'); library(coverage)"]
RUN ["Rscript", "-e", "install.packages('pytest', repo='https://cloud.r-project.org'); library(pytest)"]
RUN ["Rscript", "-e", "library(devtools); install_github('github.com/repo'); library(cool-lib)"]
COPY libs/proj libs/proj
RUN ["Rscript", "-e", "install.packages('/app/libs/proj', repos=NULL, type='source'); library(cool-proj)"]
RUN ["Rscript", "-e", "library(devtools); install_version('dtable', version='9.0', repos='http://cran.us.r-project.org'); library(dtable)"]
COPY . /app
RUN rm -rf build/
RUN rm -rf dist/
CMD /bin/bash -c \"bash build.sh --env local --log info\"\n"""

        sb.systems.generators.dockerfile.buildDockerfile(config)

        data = None
        with open(filePath, "r") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expectedDockerfile)

if __name__ == '__main__':
    unittest.main()



"""Common Global Variables"""

import pkg_resources

from colorama import Fore, Style

VERSION = pkg_resources.get_distribution("skelebot").version
DESCRIPTION = Style.BRIGHT + "{project}" + Style.RESET_ALL + """
{desc}
-----------------------------------
Version: {pVersion}
Environment: {env}
Skelebot Version: {version}
-----------------------------------"""
SKELEBOT_HOME = "~/.skelebot"
PLUGINS_HOME = "{home}/plugins".format(home=SKELEBOT_HOME)
PLUGINS_QUARANTINE = "{home}/plugins-quarantine".format(home=SKELEBOT_HOME)

LANGUAGE_IMAGE = {
    "NA": {
        "base": "ubuntu:18.04",
        "krb": "ubuntu:18.04"
    },
    "Python": {
        "base": "skelebot/python-base:{pythonVersion}",
        "krb": "skelebot/python-krb"
    },
    "R": {
        "base": "skelebot/r-base",
        "krb": "skelebot/r-krb"
    },
    "R+Python": {
        "base": "skelebot/r-base",
        "krb": "skelebot/r-krb"
    }
}
_all_deps: dict = {"Python":["numpy", "pandas", "scipy", "scikit-learn"],"R":["data.table", "here", "stringr", "readr", "testthat", "yaml"]}
LANGUAGE_DEPENDENCIES = {
    "Python": _all_deps["Python"],
    "R": _all_deps["R"],
    "R+Python": _all_deps,
}
PYTHON_VERSIONS = ['3.6', '3.7', '3.8', '3.9']

EXT_COMMAND = {"py":"python -u ", "R":"Rscript ", "sh":"bash ", "None":""}

ERROR_HEADER = Fore.RED + "ERROR" + Style.RESET_ALL + " | "
ERROR = ERROR_HEADER + "{}"
SCHEMA_ERROR = ERROR_HEADER + "skelebot.yaml | {}"

WARN_HEADER = Fore.YELLOW + "WARN" + Style.RESET_ALL + " | "
DEPRECATION_WARNING = WARN_HEADER + "The {code} has been deprecated as of v{version}. {msg}"



import boto3
from schema import Schema, And, Optional
from .artifactRepo import ArtifactRepo
from ...objects.semver import Semver

class S3Repo(ArtifactRepo):
    """
    S3 File System Class

    S3 connection details for saving artifacts in AWS S3 buckets
    """

    schema = Schema({
        'bucket': And(str, error="S3 'bucket' must be a String"),
        'region': And(str, error="S3 'region' must be a String"),
        Optional('profile'): And(str, error="S3 'profile' must be a String"),
    }, ignore_extra_keys=True)

    bucket = None
    region = None
    profile = None

    def __init__(self, bucket, region, profile):
        """ Initialize the needed values for any S3 connection """
        self.bucket = bucket
        self.region = region
        self.profile = profile

    def connect(self):
        """ Establish connection to S3 with region and profile and return the client"""
        session = boto3.Session(profile_name=self.profile, region_name=self.region)
        client = session.client('s3')
        return client

    def push(self, artifact, version, force=False, user=None, password=None):
        """ Push an artifact to S3 with the given version number """

        bucket_parts = self.bucket.split("/")
        bucket_parts.append(artifact.getVersionedName(version))
        bucket = bucket_parts[0]
        path = "/".join(bucket_parts[1:])

        # If the object exists and we are not force pushing, raise an error
        client = self.connect()
        if (not force):
            response = client.list_objects_v2(Bucket=bucket, Prefix=path)
            if ("Contents" in response.keys()):
                print(response["Contents"][0]["Key"])
                raise RuntimeError(self.ERROR_ALREADY_PUSHED)

        # Upload the artifact with the versioned name
        client.upload_file(artifact.file, bucket, path)

    def pull(self, artifact, version, currentVersion=None, override=False, user=None, password=None):
        """ Pull an artifact from S3 with the given version or the LATEST compatible version """

        # Identify the latest compatible version
        client = self.connect()
        if (version == "LATEST") and (artifact.singular == False):
            version = None
            currentSemver = Semver(currentVersion)
            ext = artifact.file.split(".")[1]
            prefix = "{name}_v{major}".format(name=artifact.name, major=currentSemver.major)

            bucket_parts = self.bucket.split("/")
            bucket = bucket_parts[0]
            bucket_parts.append(prefix)
            path = "/".join(bucket_parts[1:])
            response = client.list_objects_v2(Bucket=bucket, Prefix=path)

            # Iterate through all the artifacts int he bucket
            for content in response["Contents"]:
                artifactSemver = Semver(str(content["Key"]).split("_v")[1].split(ext)[0])

                # Find the most recent backward compatible version AKA Latest Compatible Version (LCV)
                if (currentSemver.isBackwardCompatible(artifactSemver)) and ((version is None) or (version < artifactSemver)):
                    version = artifactSemver

            # Raise an error if no LCV is found
            if (version is None):
                raise RuntimeError(self.ERROR_NOT_COMPATIBLE)

        # Download the artifact based on the version provided or LCV
        artifactName = artifact.getVersionedName(version)
        dest = artifact.file if (override) else artifactName

        bucket_parts = self.bucket.split("/")
        bucket = bucket_parts[0]
        bucket_parts.append(artifactName)
        path = "/".join(bucket_parts[1:])

        client.download_file(bucket, path, dest)



import argparse
import copy
import unittest
from unittest import mock

from schema import SchemaError

import skelebot as sb

class TestEnvironments(unittest.TestCase):

    def test_addParsers(self):
        environments = sb.components.environments.Environments()

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        subparsers = parser.add_subparsers(dest="job")
        subparsers = environments.addParsers(subparsers)

        self.assertNotEqual(subparsers.choices["envs"], None)

    @mock.patch('skelebot.components.environments.print')
    def test_execute(self, mock_print):
        config = sb.objects.config.Config()
        args = argparse.Namespace()

        environments = sb.components.environments.Environments()
        res = environments.execute(config, args)

        self.assertEqual(res, 0)
        mock_print.assert_has_calls([mock.call("[default]"), mock.call("exp")])

if __name__ == '__main__':
    unittest.main()



class Semver():
    """
    Semver Class

    Object to translate a semantic version string into an object with separate values for major, minor, and patch that allows for operations to
    be performed such as comparing two Semvers.
    """

    major = None
    minor = None
    patch = None

    def __init__(self, version):
        """ Initialize the Major, Minor, and Patch numbers of a semantic version number """
        parts = version.split(".")
        self.major = parts[0]
        self.minor = parts[1]
        self.patch = parts[2]

    def __str__(self):
        """ When converting to a String the format should read as major.minor.patch """
        return "{maj}.{min}.{pat}".format(maj=self.major, min=self.minor, pat=self.patch)

    def __lt__(self, other):
        """ Check if this semver is smaller (less recent) than another semver """
        lt = False
        if (self.major < other.major) or ((self.major == other.major) and ((self.minor < other.minor) or ((self.minor == other.minor) and (self.patch < other.patch)))):
            lt = True

        return lt

    def __eq__(self, other):
        """ Check if this semver is exactly equal to another semver """
        eq = True
        if (self.patch != other.patch) or (self.minor != other.minor) or (self.major != other.major):
            eq = False

        return eq

    def __le__(self, other):
        """ Check if this semver is larger (more recent) than another semver """
        le = False
        if (self < other) or (self == other):
            le = True

        return le

    def isBackwardCompatible(self, semver):
        """
        Determine whether this Semver is backward compatible with the Semver provided

        For this semver to be backward compatible with another the Major version of each must be
        the same and this semver must be at least as large (at least as recent) as the one provided
        """

        compat = False
        if (self.major == semver.major) and (self >= semver):
            compat = True

        return compat




"""Registry Component"""

from schema import Schema, And, Optional
from ..objects.skeleYaml import SkeleYaml
from ..objects.component import Activation, Component
from ..systems.execution import docker

class Aws(SkeleYaml):
    """
    AWS Specific Registry Details

    Provides the fields necessary for authenticating with an AWS ECR: region and profile.
    """

    schema = Schema({
        'region': And(str, error='Registry AWS \'region\' must be a String'),
        Optional('profile'): And(str, error='Registry AWS \'profile\' must be a String')
    }, ignore_extra_keys=True)

    region = None
    profile = None

    def __init__(self, region=None, profile=None):
        """Instantiate the Registry Class Object based on the provided parameters"""
        self.region = region
        self.profile = profile

class Registry(Component):
    """
    Registry Class

    Provides the ability to define the registry that will be used when publishing the Docker image
    for the project, and provides the capability to publish by logging into the registry through
    the CLI.
    """

    activation = Activation.PROJECT
    commands = ["publish"]

    schema = Schema({
        Optional('host'): And(str, error='Registry \'host\' must be a String'),
        Optional('port'): And(int, error='Registry \'port\' must be an Integer'),
        Optional('user'): And(str, error='Registry \'user\' must be a String'),
        Optional('aws'): And(dict, error='Registry \'aws\' must be a Dictionary')
    }, ignore_extra_keys=True)

    host = None
    port = None
    user = None
    aws = None

    @classmethod
    def load(cls, config):
        """Instantiate the AWS Class Object if it is present"""

        cls.validate(config)
        aws = Aws.load(config["aws"]) if ("aws" in config) else None

        return cls(config.get("host"), config.get("port"), config.get("user"), aws)

    def __init__(self, host=None, port=None, user=None, aws=None):
        """Instantiate the Registry Class Object based on the provided parameters"""
        self.host = host
        self.port = port
        self.user = user
        self.aws = aws

    def addParsers(self, subparsers):
        """
        SkeleParser Hook

        Adds a parser for the publish command to allow the user to login and push the project's
        Docker image to the defined registry, or Docker Hub, if one is not defined.
        """

        helpMessage = "Publish your versioned Docker Image to the registry"
        registryParser = subparsers.add_parser("publish", help=helpMessage)
        registryParser.add_argument("-t", "--tags", nargs='*', help="Additional image tags")

        return subparsers

    def execute(self, config, args, host=None):
        """
        Execution Hook

        Executes when the publish command is provided and prompts for username and password before
        building the project's Docker Image and pushing it to the defined registry.
        """

        # Login to the registry
        if self.aws is not None:
            docker.loginAWS(self.host, self.aws.region, self.aws.profile, docker_host=host)
        else:
            docker.login(host=self.host, docker_host=host)

        # Build and Push with the different tags (LATEST and VERSION default)
        if (not args.skip_build_global):
            docker.build(config, host=host)

        docker.push(config, self.host, self.port, self.user, tags=args.tags, docker_host=host)



from unittest import TestCase, main
import skelebot as sb

class TestArtifactRepo(TestCase):

    def test_push_pass(self):
        artifactRepo = sb.components.repository.artifactRepo.ArtifactRepo()
        nothing = artifactRepo.push("artifact", "1.1.0", force=False, user=None, password=None)
        self.assertTrue(nothing is None)

    def test_pull_pass(self):
        artifactRepo = sb.components.repository.artifactRepo.ArtifactRepo()
        nothing = artifactRepo.pull("artifact", "1.1.0", currentVersion="1.2.0", override=False, user=None, password=None)
        self.assertTrue(nothing is None)

if __name__ == '__main__':
    main()
